from datetime import datetime, timezone
import yaml
import numpy as np


def ordinal(n):
    return f"{n}{'th' if 11<=n<=13 else {1:'st',2:'nd',3:'rd'}.get(n%10, 'th')}"

def get_readable_date(date_obj=None, tz=None):
    if date_obj is None:
        date_obj = datetime.now().replace(tzinfo=timezone.utc)

    if tz:
        date_obj = date_obj.replace(tzinfo=tz)

    return date_obj.strftime(f"%a {ordinal(date_obj.day)} of %b %Y")

def inspect_dataframe(df):
    # Create a dictionary to store column statistics
    stats_dict = {}

    # Iterate over each column in the dataframe
    for column in df.columns:
        # Gather common statistics
        col_stats = {
            'dtype': str(df[column].dtype),
            'count_of_values': int(df[column].count()),
            'count_of_nulls': int(df[column].isna().sum())
        }

        # Check if the column is numerical
        if np.issubdtype(df[column].dtype, np.number):
            col_stats['mean'] = float(df[column].mean())
        
        # Special handling for datetime columns
        elif np.issubdtype(df[column].dtype, np.datetime64):
            sorted_dates = np.sort(df[column].dropna())  # Sort dates and drop NaNs
            if len(sorted_dates) > 1:
                intervals = np.diff(sorted_dates)  # Calculate differences between consecutive dates
                col_stats['interval_between_records'] = str(np.mean(intervals))

        # Handle non-numerical columns
        else:
            non_null_values = df[column].dropna()
            if not non_null_values.empty:
                col_stats['first_value'] = str(non_null_values.iloc[0])
                col_stats['last_value'] = str(non_null_values.iloc[-1])

        # Add the stats to the main dictionary, using the column name as the key
        stats_dict[column] = col_stats

    # Convert dictionary to YAML format
    return yaml.dump(stats_dict, sort_keys=False, default_flow_style=False)



import os
import time
import google.generativeai as gemini

try:
    # Attempt package-relative import
    from . import output_manager
except ImportError:
    # Fall back to script-style import
    import output_manager

output_manager = output_manager.OutputManager()

def init():
    API_KEY = os.environ.get('GEMINI_API_KEY')
    gemini.configure(api_key=API_KEY)

def convert_openai_to_gemini(messages):
    updated_data = []
    system_content = None
    for item in messages:
        if item['role'] == 'system':
            system_content = item['content']
            continue 
        if item['role'] == 'assistant':
            item['role'] = 'model'
        item['parts'] = f"[{item.pop('content').strip()}]"
        updated_data.append(item)

    return updated_data, system_content

def llm_call(messages: str,model_name: str,temperature: str,max_tokens: str):  

    init()

    messages, system_instruction = convert_openai_to_gemini(messages)

    generation_config = {
        "temperature": temperature,
        "top_p": 1,
        "top_k": 1,
        "max_output_tokens": max_tokens,
    }

    model = gemini.GenerativeModel(model_name=model_name,
        generation_config=generation_config,
        system_instruction = system_instruction
    )

    start_time = time.time()


    response = model.generate_content(messages)

    end_time = time.time()

    elapsed_time = end_time - start_time

    content = response.text
    prompt_tokens_used = model.count_tokens(messages).total_tokens
    completion_tokens_used = model.count_tokens(content).total_tokens
    total_tokens_used = prompt_tokens_used + completion_tokens_used

    if elapsed_time > 0:
        tokens_per_second = completion_tokens_used / elapsed_time
    else:
        tokens_per_second = 0

    return content, messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second

def llm_stream(log_and_call_manager, chain_id: str,messages: str,model_name: str,temperature: str,max_tokens: str,tools: str = None):
    collected_messages = []  

    init()

    messages, system_instruction = convert_openai_to_gemini(messages)

    generation_config = {
        "temperature": temperature,
        "top_p": 1,
        "top_k": 1,
        "max_output_tokens": max_tokens,
    }

    model = gemini.GenerativeModel(model_name=model_name,
        generation_config=generation_config,
        system_instruction = system_instruction,
    )

    response = model.generate_content(messages, stream=True)

    start_time = time.time()
    for chunk in response:
        if chunk.text is not None:
            chunk_message = chunk.text
            collected_messages.append(chunk_message)
            output_manager.print_wrapper(chunk_message, end='', flush=True)  

    end_time = time.time()
    elapsed_time = end_time - start_time

    output_manager.print_wrapper("")

    full_reply_content = ''.join([m for m in collected_messages])

    # Count tokens used
    completion_tokens_used = model.count_tokens(full_reply_content).total_tokens
    prompt_tokens_used = model.count_tokens(messages).total_tokens

    # calculate the total tokens used
    total_tokens_used = prompt_tokens_used + completion_tokens_used
    
    if elapsed_time > 0:
        tokens_per_second = completion_tokens_used / elapsed_time
    else:
        tokens_per_second = 0

    return full_reply_content, messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second


import hashlib
import os
from openai import OpenAI
from pinecone import Pinecone, ServerlessSpec

try:
    # Attempt package-relative import
    from . import output_manager
except ImportError:
    # Fall back to script-style import
    import output_manager

output_handler = output_manager.OutputManager()

class EmbeddingClientIntegration:
    def vectorize(self):
        raise NotImplementedError
    

class OpenAIEmbeddingClient(EmbeddingClientIntegration):
    def __init__(self):
        self.client = OpenAI()

    def vectorize(self, question):
        response = self.client.embeddings.create(
            input=question,
            model="text-embedding-3-small"
        )
        return response.data[0].embedding
    

class HFSentenceTransformersClient(EmbeddingClientIntegration):
    def __init__(self):
        try:
            from sentence_transformers import SentenceTransformer
            self.model = SentenceTransformer('all-MiniLM-L6-v2')
        except ImportError:
            raise RuntimeError("Sentence Transformers library is not installed.")

    def vectorize(self, question):
        return self.model.encode([question])[0].tolist()
    

class PineconeWrapper:
    def __init__(self):
        self.api_key = os.getenv('PINECONE_API_KEY')
        if not self.api_key:
            return
        
        self.cloud = "aws"
        self.env = "us-east-1"
        self.embed_platform = "openai"  # "openai" or "hf_sentence_transformers"
        self.pinecone_client = Pinecone(api_key=self.api_key)
        self.embedding_client = self.initialize_embedding_client()
        self.index_name, self.dimension = self.determine_index_settings()
        self.ensure_index_exists()

    def initialize_embedding_client(self):
        if self.embed_platform == "openai":
            return OpenAIEmbeddingClient()
        elif self.embed_platform == "hf_sentence_transformers":
            try:
                return HFSentenceTransformersClient()
            except RuntimeError as e:
                output_handler.display_system_messages(e)
        else:
            raise ValueError("Unsupported embedding platform")
        
    def determine_index_settings(self):
        settings = {
            "hf_sentence_transformers": ("bambooai-qa-retrieval", 384),
            "openai": ("bambooai-qa-retrieval", 1536)
        }
        return settings.get(self.embed_platform, (None, None))

    def vectorize_question(self, question):
        return self.embedding_client.vectorize(question)

    def ensure_index_exists(self):
        if self.index_name not in self.pinecone_client.list_indexes().names():
            output_handler.display_system_messages(f"Creating a new vector db index. Please wait... {self.index_name}")
            self.pinecone_client.create_index(
                name=self.index_name,
                metric="cosine",
                dimension=self.dimension,
                spec=ServerlessSpec(cloud=self.cloud, region=self.env)
            )
        self.index = self.pinecone_client.Index(name=self.index_name)

    def query_index(self, question):
        # Vectorize the question
        vectorised_question = self.vectorize_question(question)
        # Query the vector db
        results = self.index.query(vector=vectorised_question, top_k=1, include_values=True)
        matches = results.get('matches', [])

        if not matches:
            output_handler.display_system_messages("I was unable to find a matching record in the vector db.")
            return None

        return matches[0]
    
    def fetch_record(self, id):
        fetched_data = self.index.fetch(ids=[id])
        vector_data = fetched_data.get('vectors', {}).get(id, {})
        if not vector_data:
            output_handler.display_system_messages("No data found for this vector db id")
            return None
        return vector_data

    def check_similarity(self, match, similarity_threshold):
        match_id = match['id']
        similarity_score = match['score']
        output_handler.display_system_messages(f"\nClosest match vector db record: {match_id}, Similarity score: {similarity_score}\n")
        
        # Check if the similarity score is above the threshold
        if match['score'] < similarity_threshold:
            output_handler.display_system_messages(f"Similarity score {match['score']} is below the threshold {similarity_threshold}")
            return False
        
        return True 

    def retrieve_matching_record(self, question, df_columns, similarity_threshold, match_df=True,):
        match = self.query_index(question)
        if match and self.check_similarity(match,similarity_threshold):
            vector_data = self.fetch_record(match['id'])
            if not vector_data:
                return None
            # Get the metadata
            metadata = vector_data['metadata']
            # Check if the dataframe columns match
            if match_df:
                if metadata['df_col'] == df_columns:
                    return vector_data
                else:
                    output_handler.display_system_messages("The dataframe columns do not match. I will not use this record.")
                    return None
            else:
                return vector_data
        else:
            return None      

    def add_record(self, question, plan, df_columns, code, new_rank, similarity_threshold):
        new_rank = int(new_rank)

        if new_rank < 8:
            output_handler.display_system_messages("The new rank is below the threshold. Not adding/updating the vector db record.")
            return
        
        # Generate a unique id for the record.
        id = hashlib.sha256(question.encode()).hexdigest() # Placeholder

        # Vectorize the question
        vectorised_question = self.vectorize_question(question)

        # Fetch the record with closest match to the question
        vector_data = self.retrieve_matching_record(question, df_columns, similarity_threshold)

        vector_rank = -1 # Default rank if the record does not exist  

        if vector_data:
            vector_rank = vector_data['metadata']['rank']
            id = vector_data['id']
        # If the new rank is higher, add/update the record
        if vector_rank < new_rank:
            metadata = {"plan": plan, "df_col": df_columns, "question_txt": question, "code": code, "rank": new_rank}
            vectors = [(id, vectorised_question, metadata)]
            self.index.upsert(vectors=vectors)
            output_handler.display_system_messages(f"\nAdded/Updated the vector db record with id: {id}")
        else:
            output_handler.display_system_messages(f"Existing rank {vector_rank} is higher or equal to the new rank. I am not updating the existing vector db record.")



# prompts.py

# Default Examples (Otherwise Pinecone Long Term Memory)
default_example_output_df = """
Example Output:

```python
import pandas as pd

# Identify the dataframe `df`
# df has already been defined and populated with the required data

# Call the `describe()` method on `df`
df_description = df.describe()

# Print the output of the `describe()` method
print(df_description)
```
"""
default_example_output_gen = """
Example Output:

```python
# Import required libraries
import yfinance as yf
import matplotlib.pyplot as plt

# Define the ticker symbol
tickerSymbol = 'AAPL'

# Get data on this ticker
tickerData = yf.Ticker(tickerSymbol)

# Get the historical prices for this ticker
tickerDf = tickerData.history(period='1d', start='2010-1-1', end='2021-1-1')

# Normalize the data
tickerDf = tickerDf[['Close']]
tickerDf = tickerDf.reset_index()
tickerDf = tickerDf.rename(columns={'Date': 'ds', 'Close': 'y'})

# Plot the close prices
plt.plot(tickerDf.ds, tickerDf.y)
plt.show()
```
"""
default_example_plan_df = """
EXAMPLE:
Reflection on the problem
...

```yaml
plan:
  - "Step 1: Convert the 'datetime(GMT)' ..."
  - "Step 2: Calculate the total..."
  - "Step 3: Calculate the pace..."
  - ...
```
"""

default_example_plan_gen = """
EXAMPLE:
Reflection on the problem
...

```yaml
plan:
  - "Step 1: Import the yfinance......"
  - "Step 2: Define the ticker..."
  - "Step 3: Download data..."
  - ...
  ```
"""
# Expert Selector Agent Prompts
#   - An 'Internet Research Specialist' that can search the internet to find factual information, relevant data, and contextual details to help address user questions.
expert_selector_system = """
You are a classification expert, and your job is to classify the given task, and select the expert best suited to solve the task.

1. Determine whether the solution will require an access to a dataset that contains various data, related to the question.

2. Select an expert best suited to solve the task, based on the outcome of the previous step.
   The experts you have access to are as follows:

   - A 'Data Analyst' that can deal with any questions that can be directly solved with code.
   - A 'Research Specialist' that can answer questions on any subject that do not require coding, incorporating tools like Google search and LLM as needed.

3. State your level of confidence that if presented with this task, you would be able to solve it accurately and factually correctly on a scale from 0 to 10. Output a single integer.

Formulate your response as a JSON string, with 3 fields {requires_dataset (true or false}, expert, confidence}. Always enclose the JSON string within ```json tags

Example Query:
How many rows are there in this dataset ?

Example Output:
```json
{
  "requires_dataset": true,
  "expert": "Data Analyst",
  "confidence": 10
}
"""
expert_selector_user = """
The user asked the following question: '{}'.
"""
# Analyst Selector Agent Prompts
analyst_selector_system = """
You are a classification expert, and your job is to classify the given task.

1. Select an analyst best suited to solve the task.
  The analysts you have access to are as follows:

    - A 'Data Analyst DF':
      Select this expert if user provided a dataframe. The DataFrame 'df' is already defined and populated with necessary data.

    - A 'Data Analyst Generic':
      Select this expert if user did not provide the dataframe.

2. Rephrase the query.
    - Focus on incorporating previous context and ensure accuracy in spelling, syntax, and grammar. 
    - Capture every nuance of the user's request meticulously, omitting no details.
    - Place significant emphasis on the query that immediately precedes this one. 
    - The rephrased version should be both as descriptive as possible while concise, suitable for later conversion into a detailed, multi-step action plan.
    - There is no need to include the dataframe details in the rephrased query.

Formulate your response as a JSON string, with 2 fields {analyst,rephrased_query}. Always enclose the JSON string within ```json tags

Example Query:
How many rows in this dataset ?

Example Output:
```json
{
  "analyst": "Data Analyst DF",
  "rephrased_query": "How many rows does this dataframe contain?"
}
"""
analyst_selector_user = """
DATAFRAME COLUMNS:

{}

QUESTION:

{}
"""
# Theorist Agent Prompts
theorist_system = """
You are a Research Specialist and your job is to find answers and educate the user. 
Provide factual information responding directly to the user's question. Include key details and context to ensure your response comprehensively answers their query.

Today's Date is: {}

The user asked the following question: '{}'.
"""
# Planner Agent Prompts
planner_system = """
You are an AI assistant capable of assisting users with various tasks related to research, coding, and data analysis. 
The user will inform you about the expertise required to accomplish their task.
You have access to a Google search tool and can retrieve any information that might be missing.

Today's Date is: {}
"""
planner_user_df = """
TASK:
{}

DATAFRAME:

```yaml
{}
```
First: Evaluate whether you have all necessary and requested information to provide a solution. 
Use the dataset description above to determine what data and in what format you have available to you.
You are able to search internet if the user asks for it, or you require any information that you can not derive from the given dataset or the instruction.

Second: Reflect on the problem and briefly describe it, while addressing the problem goal, inputs, outputs,
rules, constraints, and other relevant details that appear in the problem description.

Third: Based on the preceding steps, formulate your response as an algorithm, breaking the solution in up to eight simple yet descriptive, clear English steps. 
You MUST Include all values or instructions as described in the above task!
If fewer steps suffice, that's acceptable. If more are needed, please include them.
Remember to explain steps rather than write code.

This algorithm will be later converted to Python code and applied to the pandas DataFrame 'df'.
The DataFrame 'df' is already defined and populated with data! 

Output the algorithm as a YAML string. Always enclose the YAML string within ```yaml tags.

Allways make sure to incorporate any details or context from the previous conversations, that might be relevant to the task at hand

{}
"""
planner_user_gen = """
TASK:
{}

First: Evaluate whether you have all necessary and requested information to provide a solution.
You are able to search internet if you require any information that you can not derive from the instruction.

Second: Reflect on the problem and briefly describe it, while addressing the problem goal, inputs, outputs,
rules, constraints, and other relevant details that appear in the problem description.

Second: Based on the preceding steps, formulate your response as an algorithm, breaking the solution in up to eight simple yet descriptive, clear English steps. 
You MUST include all values, instructions links or URLs necessary to solve the above task!
If fewer steps suffice, that's acceptable. If more are needed, please include them. 
Remember to explain steps rather than write code.
This algorithm will be later converted to Python code.

Output the algorithm as a YAML string. Always enclose the YAML string within ```yaml tags.

Allways make sure to incorporate any details or context from the previous conversations, that might be relevant to the task at hand.

{}
"""
# Code Generator Agent Prompts
code_generator_system_df = """
You are an AI data analyst and your job is to assist users with analyzing data in the pandas dataframe.
The user will provide a dataframe named `df`, and the task formulated as a list of steps to be solved using Python.
The dataframe df has already been defined and populated with the required data!

Please make sure that your output contains a FULL, COMPLETE CODE that includes all steps, and solves the task!
Always include the import statements at the top of the code.
Always include print statements to output the results of your code.
"""
code_generator_system_gen = """
You are an AI data analyst and your job is to assist users with data analysis, or any other tasks related to coding. 
You have not been provided with any datasets, but you have access to the internet.
The user will provide the task formulated as a list of steps to be solved using Python. 

Please make sure that your output contains a FULL, COMPLETE CODE that includes all steps, and solves the task!
Always include the import statements at the top of the code.
Always include print statements to output the results of your code.
"""
code_generator_user_df = """
TASK:
{}

PLAN:
```yaml
{}
```

DATAFRAME `print(df.dtypes)`:
{}

CODE EXECUTION OF THE PREVIOUS TASK RESULTED IN:
{}


{}
"""
code_generator_user_gen = """
TASK:
{}

PLAN:
``yaml
{}
```

CODE EXECUTION OF THE PREVIOUS TASK RESULTED IN:
{}


{}
"""
# Error Corrector Agent Prompts
error_corector_system = """
The execution of the code that you provided in the previous step resulted in an error.
Return a complete, corrected python code that incorporates the fixes for the error.
Always include the import statements at the top of the code, and comments and print statements where necessary.

The error message is: {}
"""
# Code Debugger Prompts
code_debugger_system = """
Your job as an AI QA engineer involves correcting and refactoring of the given Code so it delivers the outcome as described in the given Task list.

Code:
{}.
Task list:
{}.

Please follow the below instructions to accomplish your assingment.If provided, the dataframe df has already been defined and populated with the required data.

Task Inspection:
Go through the task list and the given Python code side by side.
Ensure that each task in the list is accurately addressed by a corresponding section of code. 
Do not move on to the next task until the current one is completely solved and its implementation in the code is confirmed.

Code Sectioning and Commenting:
Based on the task list, divide the Python code into sections. Each task from the list should correspond to a distinct section of code.
At the beginning of each section, insert a comment or header that clearly identifies the task that section of code addresses. 
This could look like '# Task 1: Identify the dataframe df' for example.
Ensure that the code within each section correctly and efficiently completes the task described in the comment or header for that section.

After necessary modifications, provide the final, updated code, and a brief summary of the changes you made.
Always use the backticks to enclose the code.

Example Output:
```python
import pandas as pd

# Task 1: Identify the dataframe `df`
# df has already been defined and populated with the required data

# Task 2: Call the `describe()` method on `df`
df_description = df.describe()

# Task 3: Print the output of the `describe()` method
print(df_description)
```
"""
# Code Ranker Agent Prompts
code_ranker_system = """
As an AI QA Engineer, your role is to evaluate and grade the code: {}, supplied by the AI Data Analyst. You should rank it on a scale of 1 to 10.

In your evaluation, consider factors such as the relevancy and accuracy of the obtained results: {} in relation to the original assignment: {},
clarity of the code, and the completeness and format of outputs.

For most cases, your ranks should fall within the range of 5 to 7. Only exceptionally well-crafted codes that deliver exactly as per the desired outcome should score higher. 

Please enclose your ranking in <rank></rank> tags.

Example Output:
<rank>6</rank>
"""
# Solution Summarizer Agent Prompts
solution_summarizer_system = """
The user presented you with the following question.
Question: {}

To address this, you have designed an algorithm.
Algorithm: {}.

You have crafted a Python code based on this algorithm, and the output generated by the code's execution is as follows.
Output: {}.

Please provide a summary of insights achieved through your method's implementation.
Present this information in a manner that is both clear and easy to understand.
Ensure that all results from the computations are included in your summary.
If the user asked for a particular information that is not included in the code execution results, and you know the answer please incorporate the answer to your summary.
"""
# Google Search Query Generator Agent Prompts
google_search_query_generator_system = """
You are an AI internet research specialist and your job is to formulate a user's question as a search query.
Reframe the user's question into a search query as per the below examples.

Example input: Can you please find out what is the popularity of Python programming language in 2023?
Example output: Popularity of Python programming language in 2023

The user asked the following question: '{}'.
"""
# Google Search Summarizer Agent Prompts
google_search_summarizer_system = """
Read the following text carefully to understand its content. 
  
Text:

{}

Based on your understanding, provide a clear and comprehensible answer to the question below by extracting relevant information from the text.
Be certain to incorporate all relevant facts and insights.
Fill in any information that user has asked for, and that is missing from the text.

Question: {}
"""
google_search_react_system = """
You are an Internet Research Specialist, and run in a loop of Thought, Action, Observation. This Thought, Action, Observation loop is repeated until you output an Answer.
At the end of the loop you output an Answer.
Use Thought to describe your thoughts about the question you have been asked.
Use Action to run one of the actions available to you.
Observation will be the result of running those actions.

Your available actions are:

calculate:
e.g. calculate: 4 * 7 / 3
Runs a calculation and returns the number - uses Python so be sure to use floating point syntax if necessary

google_search:
e.g. google_search: Popularity of the Python programming language in 2022
Returns a summary of a Google Search
Today's Date is: {}

Use Google Search ONLY if you dont know the answer to the question!

Example session:

Question: What is Leonardo di Caprio's girlfriends age raised to the power of 2?\n
Thought: I need to search for Leonardo DiCaprio's girlfriend's name.\n
Action: google_search: Leonardo DiCaprio's girlfriend's name\n

You will be called again with this:

Observation: Leonardo DiCaprio has had a string of high-profile relationships over the years, including with models Gisele Bündchen, Bar Refaeli, and Nina Agdal. As of 2023, he is currently dating actress and model Camila Morrone.

You then output:

Thought: Camila Morrone's age.
Action: google_search: Camila Morrone's age

You will be called again with this:

Observation: Camila Morrone is 23 years old.

You then output:

Thought: Camila Morrone is 23 years old. I need to raise 23 to the power of 2.
Action: calculate: 23**2

You will be called again with this:

Observation: 529

You then output the finall answer:

Answer: Leonardo's current girlfriend is Camila Morrone, who is 23 years old. 23 raised to the power of 2 is 529.
"""


from .bambooai import BambooAI
from . import prompts
from . import func_calls
from . import qa_retrieval
from . import models
from . import google_search
from . import reg_ex
from . import local_models
from . import groq_models
from . import openai_models
from . import ollama_models
from . import gemini_models
from . import anthropic_models
from . import mistral_models
from . import log_manager
from . import output_manager
from . import utils


from setuptools import setup, find_packages
import os

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='bambooai',
    version='0.3.49',
    description='A lightweight library for working with pandas dataframes using natural language queries',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Palo Galko',
    packages=find_packages(),
    install_requires=[
        'openai',
        'tiktoken',
        'pandas',
        'termcolor',
        'newspaper3k',
        'pinecone-client',
        'sentence-transformers',
        'google-generativeai',
        'groq',
        'ollama',
        'anthropic',
        'mistralai'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
)